'''
    Copyright 2018 University of Michigan
 
    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
'''
from datetime import date, datetime, timedelta
import decimal
import numpy as np
import pandas as pd


class SingleResultSet:
    """Reads the data from the Java SingleResultSet

    The data retrieval roughly follows PEP 249:
    https://www.python.org/dev/peps/pep-0249/#type-objects-and-constructors
    """

    python_type_to_numpy_type = {
        bool: np.bool_,
        int: np.int64,
        decimal.Decimal: np.float64,
        str: np.unicode_,
        bytes: np.unicode_,
        date: np.datetime64,
        datetime: np.datetime64,
        timedelta: np.timedelta64
    }

    def __init__(self, heading, column_types, rows, verdict_context):
        self._verdict_context = verdict_context
        # (heading, column_inttypes, column_types, rows) = self._read_all(resultset)
        self._heading = heading
        self._column_types = column_types
        self._rows = rows
        self.rowcount = len(rows)

    @classmethod
    def from_java_resultset(cls, resultset, verdict_context):
        (heading, column_inttypes, column_types, rows) = cls._read_all(resultset, verdict_context)
        return cls(heading, column_types, rows, verdict_context)

    @classmethod
    def status_result(cls, status_msg, verdict_context):
        heading = ['status']
        column_types = ['string']
        rows = [[status_msg]]
        return cls(heading, column_types, rows, verdict_context)

    def column_names(self):
        return self._heading

    def rows(self):
        return self._rows

    def fetchall(self):
        return self.rows()

    def types(self):
        return self._column_types

    # def type_java_int(self):
    #     return self._column_inttypes

    def to_df(self):
        """
        Converts to a pandas DataFrame.
        """
        rows = self.rows()
        columns = self.column_names()
        if len(columns) == 0 and len(rows) == 0:
            columns = ['empty']
        return pd.DataFrame(np.array(rows), columns=columns)

    @classmethod
    def _read_value(cls, resultset, index, col_type, verdict_context):
        dbtype = verdict_context.get_dbtype()
        if dbtype == 'mysql':
            return cls._read_value_mysql(resultset, index, col_type)
        elif dbtype == 'presto':
            return cls._read_value_presto(resultset, index, col_type)
        else:
            raise NotImplementedError

    @classmethod
    def _read_value_presto(cls, resultset, index, col_type):
        """
        Type conversion rule (Presto):
        'tinyint'    => int
        'boolean'    => int
        'smallint'   => int
        'integer'    => int
        'bigint'     => int
        'decimal'    => float
        'real'       => float
        'double'     => float
        'date'       => str
        'timestamp'  => str
        'char'       => str
        'varchar'    => str
        """
        type_to_read_in_str_for_presto = set(['decimal', 'date', 'timestamp'])

        if col_type in type_to_read_in_str_for_presto:
            value_str = resultset.getString(index)
            if value_str is None:
                return None

            if col_type == 'decimal':
                return float(value_str)
            elif col_type == 'date':
                return value_str
            elif col_type == 'timestamp':
                return value_str[:17] + '%06.3f'%float(value_str[18:])
            else:
                return None     # not supposed to reach here

        return resultset.getValue(index)

    @classmethod
    def _read_value_mysql(cls, resultset, index, col_type):
        """Reads the value in a type-sensitive way

        Due to the reliance on the underlying JDBC library, we cannot always distinguish the exact
        type of the underlying data; we can only set the type according to the type indentifier as
        indicated by the JDBC library itself. For example, MySQL's library does not distinguish bit and
        boolean by passing -7 for both types while their actual meanings are different and they are
        separate types defined in java.sql.Types.

        Type conversion rule (MySQL):
        'bit'                     => boolean,   # due to MySQL driver bug ('bit' is treated as boolean)
        'tinyint'                 => int,
        'bool'                    => boolean,
        'smallint'                => int,
        'medimumInteger'          => int,
        'int'                     => int,
        'integer                  => int,
        'bigint'                  => int,
        'decimal'                 => decimal.Decimal,
        'dec'                     => decimal.Decimal,
        'real'                    => float,
        'double'                  => float,
        'doubleprecision'         => float,
        'date'      => JavaObject => datetime.date,
        'datetime'  => JavaObject => datetime.datetime,
        'timestamp' => JavaObject => datetime.datetime,
        'time'      => JavaObject => datetime.timedelta,
        'year(2)'   => JavaObject => datetime.date,
        'year(4)'   => JavaObject => datetime.date,
        'char'                    => str,
        'varchar'                 => str,
        'binary'                  => bytes,
        'varbinary'               => bytes,
        'tinyblob'                => bytes,
        'tinytext'                => str,
        'blob'                    => bytes,
        'text'                    => str,
        'mediumBlob'              => bytes,
        'medimumText'             => str,
        'longBlob'                => bytes,
        'longText'                => str,
        'enumCol'                 => str,
        'setCol'                  => str

        Time-related Java objects are read as py4j.JavaObject by default. To avoid this, we read
        them as str and convert to an appropriate python object.

        Args
            resultset: the Java result set currently in process
            index: zero-based index of the column to read
            col_type: column type in str
        """
        type_to_read_in_str_for_mysql = set(['date', 'timestamp', 'time', 'bigint'])

        if col_type in type_to_read_in_str_for_mysql:
            value_str = resultset.getString(index)
            if value_str is None:
                return None

            if col_type == 'date':
                return datetime.strptime(value_str, "%Y-%m-%d").date()
            elif col_type == 'timestamp':
                value_str = value_str[:19]      # 19 == len('2018-12-31 00:00:01')
                return datetime.strptime(value_str, "%Y-%m-%d %H:%M:%S")
            elif col_type == 'time':
                t = datetime.strptime(value_str, "%H:%M:%S")
                return timedelta(hours=t.hour, minutes=t.minute, seconds=t.second)
            elif col_type == 'bigint':
                return int(value_str)
            else:
                return None         # not supposed to reach here
        # if col_type == 'bit':
        #     value = resultset.getByte(index)
        #     return value        # this will return int 0 or int 1
        else:
            return resultset.getValue(index)

    @classmethod
    def _read_all(cls, resultset, verdict_context):
        column_count = resultset.getColumnCount()
        heading = []          # column heading
        column_inttypes = []  # column types in java.sql.Types
        column_types = []     # column type
        rows = []             # data in the table

        for i in range(column_count):
            heading.append(resultset.getColumnName(i))
            column_inttypes.append(resultset.getColumnType(i))
            column_types.append(resultset.getColumnTypeName(i))

        while (resultset.next()):
            row = []
            for i in range(column_count):
                row.append(cls._read_value(resultset, i, column_types[i], verdict_context))
            rows.append(row)

        return (heading, column_inttypes, column_types, rows)



'''
    Copyright 2018 University of Michigan
 
    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
'''
import json
import os
import pkg_resources
from email import message_from_string

metadata = {}

def load_metadata():
    global metadata
    root_dir = os.path.dirname(os.path.abspath(__file__))
    metadata_filename = os.path.join(root_dir, 'metadata.json')
    metadata = json.load(open(metadata_filename))

def get_metadata(name):
    load_metadata()
    return metadata[name]

def get_verdictdb_version():
    '''
    Relies on the package metadata. See below for how to read metadata.
    https://stackoverflow.com/questions/10567174/how-can-i-get-the-author-name-project-description-etc-from-a-distribution-objec
    '''
    # pkgInfo = pkg_resources.get_distribution('pyverdict').get_metadata('PKG-INFO')
    # msg = message_from_string(pkgInfo)
    # metadata = json.loads(msg['Description'])
    # verdictdb_version = metadata['Internal VerdictDB version']
    return metadata['__verdictdb_version__']



'''
    Copyright 2018 University of Michigan
 
    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
'''
import json
import setuptools
import os
import re
import subprocess
from build_lib import build_and_copy

pyverdict_version = '0.1.2.15'


def get_verdict_jar(lib_dir):
    lib_files = os.listdir(lib_dir)
    print(lib_files)

    def get_version(libfile):
        reobj = re.search('verdictdb-core-(.*?)-jar-with-dependencies.jar', libfile)
        if reobj is None:
            return (None, None)
        else:
            version = reobj.group(1)
            return (libfile, version)

    for libfile in lib_files:
        (libfile, version) = get_version(libfile)
        if version is not None:
            return (libfile, version)
    return (None, None)

root_dir = os.path.dirname(os.path.abspath(__file__))
lib_dir = os.path.join(root_dir, 'pyverdict', 'verdict_jar')

# if the directory does not exist, we create the directory, build a jar file, and copy the jar file
# to the lib directory.
if not os.path.exists(lib_dir):
    build_and_copy(root_dir, lib_dir)

(verdict_jar_file, verdictdb_version) = get_verdict_jar(lib_dir)
if verdictdb_version is None:
    build_and_copy(root_dir, lib_dir)
    (verdict_jar_file, verdictdb_version) = get_verdict_jar(lib_dir)


# creates a metadata file
metadata_filename = os.path.join(root_dir, 'pyverdict', 'metadata.json')
metadata_file = open(metadata_filename, 'w')
json.dump(
    {
        '__version__': pyverdict_version,
        '__verdictdb_version__': verdictdb_version
    },
    metadata_file)
metadata_file.close()


# the standard setup script
setuptools.setup(
    name='pyverdict',
    version=pyverdict_version,
    description='Python interface for VerdictDB',
    url='http://verdictdb.org',
    author='Barzan Mozafari, Yongjoo Park',
    author_email='mozafari@umich.edu, pyongjoo@umich.edu',
    license='Apache License, Version 2.0',
    packages=setuptools.find_packages(),
    package_data={'pyverdict': ['lib/*.jar', 'verdict_jar/*.jar']},
    include_package_data=True,
    install_requires=[
        'py4j >= 0.10.7',
        'numpy >= 1.9',
        'pandas >= 0.23'
    ]
)



'''
    Copyright 2018 University of Michigan
 
    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
'''
import atexit
import os
import pkg_resources
import sys
from .verdictresult import SingleResultSet
from . import verdictcommon
from .verdictexception import *
from py4j.java_gateway import JavaGateway
from time import sleep, time

# To properly close all connections
created_verdict_contexts = []

def close_verdict_contexts():
    for c in created_verdict_contexts:
        c.close()

atexit.register(close_verdict_contexts)


class VerdictContext:
    """
    The main Python interface to VerdictDB's Java core.

    All necessary JDBC drivers (i.e., jar files) are already included in the 
    pyverdict package. These drivers are included in the classpath.

    JVM is started on a separate process, which is to prevent KeyboardInterrupt
    (or Ctrl+C) from killing the JVM process.

    Args:
        url: jdbc connection string
        extra_class_path: The extra classpath used in addition to verdictdb's 
                          jar file. This arg can either be a single str or a
                          list of str; each str is an absolute path
                          to a jar file.
    """

    def __init__(self, url, extra_class_path=None):
        self._gateway = self._get_gateway(extra_class_path)
        self._context = self._get_context(self._gateway, url)
        self._dbtype = self._get_dbtype(url)
        self._url = url

    def close(self):
        self._context.close()
        self._gateway.close()

    @classmethod
    def new_mysql_context(cls, host, user, password=None, port=3306):
        if password is None:
            connection_string = \
                f'jdbc:mysql://{host}:{port}?user={user}'
        else:
            connection_string = \
                f'jdbc:mysql://{host}:{port}?user={user}&password={password}'
        ins = cls(connection_string)
        created_verdict_contexts.append(ins)
        return ins

    @classmethod
    def new_presto_context(cls, host, catalog, user, password=None, port=8081):
        if password is None:
            connection_string = \
                f'jdbc:presto://{host}:{port}/{catalog}?user={user}'
        else:
            connection_string = \
                f'jdbc:presto://{host}:{port}/{catalog}?' \
                f'user={user}&password={password}'
        ins = cls(connection_string)
        created_verdict_contexts.append(ins)
        return ins

    def set_loglevel(self, level):
        self._context.setLoglevel(level)

    def sql(self, query):
        return self.sql_raw_result(query).to_df()

    def sql_raw_result(self, query):
        '''
        Development API
        '''
        start_time = time()

        java_resultset = self._context.sql(query)
        if java_resultset is None:
            msg = 'processed'
            result_set = SingleResultSet.status_result(msg, self)
        else:
            result_set = SingleResultSet.from_java_resultset(java_resultset, self)

        elapsed_time = time() - start_time
        if elapsed_time < 60.0:
            elapsed_time_str = "{0:.3f} seconds".format(elapsed_time)
        else:
            elapsed_min = elapsed_time // 60
            elapsed_sec = elapsed_time % 60
            elapsed_time_str = "{0} mins {1:.3f} seconds".format(elapsed_min, elapsed_sec)
        print("{} row(s) in the result ({})".format(len(result_set.rows()), elapsed_time_str))

        return result_set

    def get_dbtype(self):
        return self._dbtype.lower()

    def _get_dbtype(self, url):
        tokenized_url = url.split(':')
        if tokenized_url[0] != 'jdbc':
            raise VerdictException('The url must start with \'jdbc\'')
        if len(tokenized_url) < 2:
            raise VerdictException(
                'This url does not seem to have valid ' \
                f'connection information: {url}')
        return tokenized_url[1]

    def _get_gateway(self, extra_class_path):
        """
        Initializes a py4j gateway.

        Args:
            class_path: Either a single str or a list of str; each str is an
                        absolute path to the jar file.
        """
        class_path = self._get_class_path(extra_class_path)
        gateway = JavaGateway.launch_gateway(
            classpath=class_path, die_on_exit=True,
            redirect_stdout=sys.stdout,
            redirect_stderr=sys.stderr,
            create_new_process_group=True)
        sleep(1)
        return gateway

    def _get_class_path(self, extra_class_path):
        """
        Returns str class path including the one for verdict jar path
        """
        lib_jar_path = self._get_lib_jars_path()
        verdict_jar_path = self._get_verdict_jar_path()

        if not os.path.isfile(verdict_jar_path):
            raise VerdictException("VerdictDB's jar file is not found.")

        str_class_path = f'{lib_jar_path}:{verdict_jar_path}'

        if extra_class_path is None:
            pass
        if isinstance(extra_class_path, str):
            str_class_path = f'{extra_class_path}:{lib_jar_path}'
        elif isinstance(extra_class_path, list):
            extra_class_path_str = ':'.join(extra_class_path)
            str_class_path = f'{extra_class_path_str}:{lib_jar_path}'

        return str_class_path

    def _get_lib_jars_path(self):
        root_dir = os.path.dirname(os.path.abspath(__file__))
        lib_dir = os.path.join(root_dir, 'lib')
        full_paths = []
        for filename in os.listdir(lib_dir):
            if filename[-3:] == 'jar':
                full_path = os.path.join(lib_dir, filename)
                full_paths.append(full_path)
        return ':'.join(full_paths)

    def _get_verdict_jar_path(self):
        root_dir = os.path.dirname(os.path.abspath(__file__))
        lib_dir = os.path.join(root_dir, 'verdict_jar')
        version = self._get_verdictdb_version();
        verdict_jar_file = f'verdictdb-core-{version}-jar-with-dependencies.jar'
        verdict_jar_path = os.path.join(lib_dir, verdict_jar_file)
        return verdict_jar_path

    def _get_verdictdb_version(self):
        return verdictcommon.get_verdictdb_version()

    def _get_context(self, gateway, url):
        return gateway.jvm.org.verdictdb.VerdictContext.fromConnectionString(url)



'''
    Copyright 2018 University of Michigan
 
    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
'''
from .verdictcontext import VerdictContext
from .verdictcommon import get_metadata
import json

__version__ = get_metadata('__version__')

__verdictdb_version__ = get_metadata('__verdictdb_version__')

def mysql_context(host, user, password=None, port=3306):
    return VerdictContext.new_mysql_context(host, user, password, port)

def mysql(host, user, password=None, port=3306):
    return mysql_context(host, user, password, port)

def presto_context(host, catalog, user, password=None, port=8081):
    return VerdictContext.new_presto_context(host, catalog, user, password, port)

def presto(host, catalog, user, password=None, port=8081):
    return presto_context(host, catalog, user, password, port)
